//ATLANT 1.0 for Cosmological Recombination: 
//Advanced Three Level Approximation for Numerical Treatment 
//of Cosmological Recombination
//13 November 2010

//This is a public code developed by 
//Department of Theoretical Astrophysics of Ioffe Physical-Technical Institute
//194021, Politekhnicheskaya 26, St.-Petersburg, Russian Federation

//Programmed by Evgeny E. Kholupenko, Alexander V. Ivanchik, 
//Sergey A. Balashev, and Dmitry A. Varshalovich

//This code can be compiled with g++ 3.2.1, "g++ -O3 atlant.cpp"
//Other compilers may demand additional header-files.

//--Including required header-files: begin-------------------

# include <iostream.h>
# include <fstream.h>
# include <math.h>
# include <stdio.h>
//--Including required header-files: end-------------------

//--Standard values of physical constants (cgs-units): begin---------------

const double G_stand=6.67428e-8;        //--Gravitational constant-----------
const double h_stand=1.054571628e-27; //--!Planck constant with bar!---------
const double c_stand=2.99792458e10;   //--speed of light---------------------
const double pi_stand=3.14159265359;
const double me_stand=9.10938215e-28; //--electron mass----------------------
const double mp_stand=1.672621637e-24; //--proton mass------------------------
const double mH_stand=mp_stand+me_stand; //--hydrogen atom mass--------------------
const double me_reduced_stand=me_stand/(1.+me_stand/mp_stand);
const double e_stand=4.80320427e-10;     //--electron charge--------------------
const double kB_stand=1.3806504e-16;  //--Boltzmann constant-----------------
const double alpha_fine_stand=e_stand*e_stand/h_stand/c_stand;    //--fine structure constant--
const double I_stand=me_reduced_stand*pow(e_stand,4)/h_stand/h_stand/2.;  //--binding energy of ground state of hydrogen atom
const double freq1_stand=I_stand/2./pi_stand/h_stand; //--hydrogen ionization frequency--
const double T_ion_stand=I_stand/kB_stand; //--hydrogen ionization energy in temperature units--
const double a_St_B_stand=pi_stand*pi_stand*pow(kB_stand,4) //--Stefan-Boltzmann constant--
             /60./c_stand/c_stand/pow(h_stand,3);  //--Stefan-Boltzmann constant--
const double sigma_t_stand=(8.*pi_stand/3)*                  //--Thomson cross section--
             pow(e_stand,4)/pow(c_stand,4)/pow(me_stand,2);  //--Thomson cross section--
const double aB_2_stand=pow(h_stand,4)/pow(e_stand,4)/pow(me_stand,2); //--square of Bohr radius--
//--Standard values of physical constants (cgs-units): end-----------------

//--Declarations of variables for used physical constants: begin-------------

double G;
double h; //--!Planck constant with bar!-------------
double c;
double pi;
double me;
double mp;
double mH;
double me_reduced;
double e;
double kB;
double alpha_fine;
double I;
double freq1;
double T_ion;
double a_St_B;
double sigma_t;
double aB_2;
//--Declarations of variables for used physical constants: end---------------

//--Main Variations of physical values due to changes of fundamental constants: begin----
double V_energy;
double V_one_photon, V_two_photon; //--coefficients for taking into account changes of transition probabilities
double V_recombination;
//--Main Variations of physical values due to changes of fundamental constants: end----

//--Declarations of assistant variables and cosmological parameters: begin----

double eV_erg;
double km_Mp;          //-- km per Mpc------------------------
double _Mpc;           //-- cm per Mpc------------------------
double H0;             //-- Hubble constant at present epoch---------
double omega_m;        //-- non-relativistic matter density----------
double omega_nu_rel;   //-- relativistic contribution of neutrinos to matter density
int nu_number;         //-- number of sorts of neutrino--------------
double omega_b;        //-- baryonic matter density---------
double ro_crit;        //-- critical density of the Universe in cgs-units----
double X_M, Y_M;       //-- hydrogen mass fraction, helium mass fraction-----
double X_N, Y_N;       //-- hydrogen particle fraction, helium particle fraction--
double He_to_H;        //-- helium to hydrogen atom mass ratio
double N0_b;           //-- baryon concentration----------------
double N0_nuclear;     //-- total concentration of nuclei-------
double N0_H;           //-- total concentration of hydrogen atoms and ions---
double N0_e;           //-- total concentration of electrons (bound and free)
double T0;             //-- CMBR temperature at present epoch------
double omega_gamma;    //-- CMBR density------------
double omega_k;        //-- Curvature density-------
double omega_lambda;   //-- Lambda density----------
double delta_el;          //-- Assistant variable 1./(X_N+2.*Y_N)
//--Declarations of assistant variables and cosmological parameters: end------

//--Variables for atomic data: begin----------------------------------
  //-- Ionization energies, erg----------------------
double I2s_HydI, I1_HydI; //--ionization energies of states of hydrogen atom---
double I2s_HeII, I1_HeII; //--ionization energies of states of helium ion------
double I2p_HeI, I2s_HeI, I1_HeI, E_2p2sHeI; //--ionization energies of states of para-helium atom---
double I2p_orto_HeI, I2s_orto_HeI, E_2p2sHeI_orto; //--ionization energies of states of ortho-helium atom---
  
  //-- Coefficients of spontaneous transitions, s^{-1}
double A_2s1sHydI;
double A_2s1sHeI;
double A_2s1sHeII;
double A_2p1s_ortoHeI;
double A_2p1s_paraHeI;

//--Helium atomic variables: begin-----------------
//--Energy levels, Statistical weights, and Transition Probabilities--
//--Ortho Structure--
double E_HeI_ortho_2s_1;
double g_HeI_ortho_2s_1;
double E_HeI_ortho_2p_1;
double g_HeI_ortho_2p_1, g_HeI_ortho_2p;
double E_HeI_ortho_3s_1;
double g_HeI_ortho_3s_1;
double E_HeI_ortho_3d_1;
double g_HeI_ortho_3d_1;
double E_HeI_ortho_3d_2;
double g_HeI_ortho_3d_2;
double A_HeI_ortho_2p_1_2s_1;
double A_HeI_ortho_3s_1_2p_1;
double A_HeI_ortho_3d_2_2p_1;
double A_HeI_ortho_3d_1_2p_1;

//--Para Structure--
double E_HeI_para_1s_0;
double g_HeI_para_1s_0;
double E_HeI_para_2s_0;
double g_HeI_para_2s_0;
double E_HeI_para_2p_1;
double g_HeI_para_2p_1;
double E_HeI_para_3s_0;
double g_HeI_para_3s_0;
double E_HeI_para_3d_2;
double g_HeI_para_3d_2;
double E_HeI_para_4s_0;
double g_HeI_para_4s_0;
double E_HeI_para_4d_2;
double g_HeI_para_4d_2;
double E_HeI_para_5s_0;
double g_HeI_para_5s_0;
double E_HeI_para_5d_2;
double g_HeI_para_5d_2;
double A_HeI_para_2p_1_1s_0;
double A_HeI_para_2p_1_2s_0;
double A_HeI_para_3s_0_2p_1;
double A_HeI_para_3d_2_2p_1;
double A_HeI_para_4s_0_2p_1;
double A_HeI_para_4d_2_2p_1;
double A_HeI_para_5s_0_2p_1;
double A_HeI_para_5d_2_2p_1;
 
double A_HeI_ortho_2p_1_para_1s_0; //--ortho-para transition probability--
//--Helium atomic variables: end------------------

  //-- transition wavelength, cm-------
double lambd_aHeII; //--for Ly-aplha transition in helium ion-----
double lambd_aHeI; //--for 2p1s transition in singlet structure of helium
double lambd_a_ortoHeI; //--for 2p1s transition in triplet structure of helium
double lambd_aHydI; //--for Ly-aplha transition in hydrogen atom--
//--Variables for atomic data: end----------------------------------

//--Reaction rate variables: begin----------------------------------

    double recomb_HeIII_var,ion_HeII_var; 
    double recomb_HeIII_var_, ion_HeII_var_;

    double recomb_HeII_var, ion_HeI_var;
    double recomb_HeII_var_, ion_HeI_var_;

    double recomb_orto_HeII_var,ion_orto_HeI_var;
    double recomb_orto_HeII_var_, ion_orto_HeI_var_;

//--total recombination coefficient for Case B of hydrogen--
//--and corresponding ionization coefficient---
    double recomb_HydII_var,ion_HydI_var;  
    double recomb_HydII_var_, ion_HydI_var_;

    double F_H=1.; //--Hydrogen Fudge Factor---------------
//--Parameters of hydrogen fudge function---------
    double A_perturbation_H=0.;
    double z_perturbation;
    double Delta_z_perturbation;

//--Reaction rate variables: end----------------------------------
//--Inhibition Coefficients: begin-----------------------
    double CHeIII=1., CHeII=1., CHeII_orto=1., CHydII=1.;
    double CHeIII_=1., CHeII_=1., CHeII_orto_=1., CHydII_=1.;
//--Inhibition Coefficients: end-----------------------
//--Additional Kinetic Coefficients: begin-------------
    double Hydrogen_FeedBack_Correction_var;
//--Additional Kinetic Coefficients: end-------------

//--Temperature treatment wizards: begin----------------
    double epsT0, epsT0_;
    double T_matter_var, T_matter_var_;
//--Temperature treatment wizards: end----------------

//--Technical variables: begin-------------------------------------

const double epsilon0=1.e-128; //--Epsilon to avoid dividing by zero--

int number_of_info_points=1; //--every "number_of_info_points"-th point---- 
                             //--will be indicated in resulting data-file--
//--redshifts of begin and end of range of calculations--
//--integration steps: absolute and relative------
double z_begin, z_end; 
double z, z_; 
double ln_1_plus_z, D_ln_1_plus_z;
double norm_var; //--norm of convergence at every step to stop calculation --

//--weights of integration method-----
double ww=0.5;
double ww_=1.-ww;

const int quantity_speices=5; //--Number of speices: HI, HII, HeI, HeII, HeIII--
double matrix[quantity_speices][quantity_speices]; //--Reaction matrix--
double matrix_[quantity_speices][quantity_speices]; //--Reaction matrix from data of previous step--
int Number_of_Iterations=1; //--Number of iterations for taking into account feedbacks
int Number_of_levels_for_Resonant_Feedback;
//--Technical variables: end---------------------------------------

//--Math functions: begin-----------------------------

double Hav(double x)
{
        double res=0.0;
        if(x>=0.0)
                res=1.;
        return res;
}
//--Math functions: end-------------------------------

//--Functions for input of initial data: begin--------------------------

//--function for shifting descriptor from current position to --
//--symbol stop_signal
int descriptor_shift(FILE* fin,char stop_signal)
{
    char var='a';
    int stream_signal=1;
    int count_shift=0;
    while((var!=stop_signal)&&stream_signal&&(count_shift<=1000))
    {
       stream_signal=fscanf(fin,"%c",&var);
       count_shift++;
    }
    if(stream_signal&&(count_shift==1001))
    {
       cout<<"file reading message: stop_signal '"<<stop_signal<<"'is lost\n";
       return 1;
    }
    return 0;
}

//--Input of fundamental constants to be used in calculations----
int init_fund_const()
{
    FILE *fin;
    if((fin=fopen("fund.in","rb+"))==NULL)
    {
         cout<<"error of opening file fund.in\n";
         return 1;
    }

    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&G);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&h);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&c);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&me);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&mp);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&e);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg",&kB);
    fclose(fin);
    pi=pi_stand;
    if((G<=0.)||(h<=0.)||(c<=0.)||(me<=0.)||(mp<=0.)||(e<=0.))
    {
        cout<<"G, h, c, m_e, m_p, e, kB values have to be positive and double";
        return 1;
    }
    return 0;
}

//--Input of cosmological constants to be used in calculations----
int init_cosmological_const()
{
    FILE *fin;
    if((fin=fopen("cosm_p.in","rb+"))==NULL)
    {
      cout<<"error of opening file cosm_p.in\n";
      return 1;
    }
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&H0);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&omega_m);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&omega_lambda);
    descriptor_shift(fin,'=');
    fscanf(fin,"%d\n",&nu_number);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&omega_b);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&X_M);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&T0);
    fclose(fin);

if((H0<=0.)||(omega_m<=0.)||(omega_lambda<=0.)||(nu_number<=0)||(omega_b<=0.)||(X_M<=0.)||(T0<=0.))
{
cout<<"Value nu_number have to be positive and integer.\n";
cout<<"Value m_nu have to be positive or equal zero and double.\n";
cout<<"Values H0, omega_m, omega_lambda, omega_b, X_M, T0 have to be positive and double.";
return 1;
}

eV_erg=1.6021765e-12;
km_Mp=3.2407765e-20;
_Mpc=3.2407765e-25;
H0*=km_Mp;
He_to_H=3.9715; //--helium to hydrogen atom mass ratio
ro_crit=3.*H0*H0/8./pi/G; //--critical density in cgs--
Y_M=1.-X_M; //--helium mass fraction--
X_N=X_M/(X_M+Y_M/He_to_H); //--hydrogen particle fraction--
Y_N=Y_M/He_to_H/(X_M+Y_M/He_to_H); //--helium particle fraction--
delta_el=1./(X_N+2.*Y_N);  //Value for connecting number of free electrons and ions--
N0_b=omega_b*ro_crit/mH; //--concentration of baryons in the present epoch--
N0_nuclear=N0_b*(X_M+Y_M/He_to_H); //--total concentration of nuclei in the present epoch--
N0_H=N0_b*X_M; //--total concentration of hydrogen nuclei in the present epoch--
N0_e=N0_b*(X_M+2.*Y_M/He_to_H); //--total concentration of electrons in the present epoch--
omega_gamma=4.*a_St_B*T0*T0*T0*T0/c/c/c/ro_crit; //--relative cmbr density in the present day--
omega_nu_rel=(7.*((double)nu_number)/8.)*pow((4./11.),(4./3.))*omega_gamma; //--relative density of relativistic neutrinos--
omega_k=1.-omega_lambda-omega_m-omega_gamma-omega_nu_rel; //--relative density of curvature energy--
   return 0;
}

int init_redshift_range()
{
    FILE *fin;
    if((fin=fopen("redshifts.in","rb+"))==NULL)
    {
         cout<<"error of opening file redshifts.in\n";
         return 1;
    }

    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&z_begin);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&z_end);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&D_ln_1_plus_z);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&norm_var);
    descriptor_shift(fin,'=');
    fscanf(fin,"%d\n",&number_of_info_points);
    fclose(fin);
    if((z_begin<=0.)||(z_end<=0.)||(z_begin<=z_end)||(number_of_info_points<1))
    {
        cout<<"redshifts should be positive, double, and z_begin>=z_end\n";
        cout<<"number_of_info_points should be positive and integer";
        return 1;
    }
    if((D_ln_1_plus_z<=0.)||(norm_var<=0.))
    {
        cout<<"absolute value of integration step and convergence norm should be positive and double\n";
        return 1;
    }
    D_ln_1_plus_z*=-1.;

    if((z_begin>5000.)&&(z_begin<7000.))
    {
        cout<<"z_begin should be larger than 7000 (i.e. including HeIII->HeII recombination)\n";
        cout<<"or less than 5000 (i.e. excluding HeIII->HeII recombination)\n";
        return 1;
    }
    if(z_begin<3000.)
    {
        cout<<"z_begin should be larger than 3000 (i.e. exclusion of HeIII->HeII recombination is impossible)\n";
        return 1;
    }
    return 0;
}

int init_fine_effects()
{
    FILE *fin;
    if((fin=fopen("fine_effects.in","rb+"))==NULL)
    {
         cout<<"error of opening file fine_effects.in\n";
         return 1;
    }

    descriptor_shift(fin,'=');
    fscanf(fin,"%d\n",&Number_of_levels_for_Resonant_Feedback);
    fclose(fin);
    if(Number_of_levels_for_Resonant_Feedback<=0)
    {
        cout<<"Number of levels for resonant feedback should be integer and positive";
        return 1;
    }
    if(Number_of_levels_for_Resonant_Feedback>1)
    {
        Number_of_Iterations=2;
    }
    
    return 0;
}

int init_fudge_factors()
{
    FILE *fin;
    if((fin=fopen("fudge_factors.in","rb+"))==NULL)
    {
         cout<<"error of opening file fudge_factors.in\n";
         return 1;
    }

    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&F_H);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&A_perturbation_H);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&z_perturbation);
    descriptor_shift(fin,'=');
    fscanf(fin,"%lg\n",&Delta_z_perturbation);
    fclose(fin);
    if((F_H<=0.)||(z_perturbation<=0.)||(Delta_z_perturbation<=0.))
    {
        cout<<"F_H should be positive and double";
        return 1;
    }
    
    return 0;
}

//--Functions for input of initial data: end--------------------------

//--Cosmological laws: begin---------------------------------------------

//Current value of Hublle constant---------------------------------------
double H(double z)
{
        return H0*sqrt(omega_lambda+omega_k*pow((1.+z),2)+omega_m*pow((1.+z),3)
                       +(omega_nu_rel+omega_gamma)*pow((1.+z),4));
}

double redshift(double ln_1_plus_z)
{
      return (exp(ln_1_plus_z)-1.);
}
//concentration change-----------------------------------------------
double N_b(double z) //--total concentration of baryons--
{
        return N0_b*(1.+z)*(1.+z)*(1.+z);
}

double N_nuclear(double z) //--total concentration of nuclei--
{
        return N0_nuclear*(1.+z)*(1.+z)*(1.+z);
}

double N_e_tot(double z) //--total concentration of electrons--
{
        return N0_e*(1.+z)*(1.+z)*(1.+z);
}

double N_H(double z) //--total concentration of hydrogen atoms and ions--
{
        return N0_H*(1.+z)*(1.+z)*(1.+z);
}

//temperature change-------------------------------------------------
double T_br(double z) //--cmbr temperature--
{
        return T0*(1.+z);
}
//--Cosmological laws: end---------------------------------------------


//--Class for description of plasma component: begin---------------------

  //--Class for ion component: begin------------------------
class comp_of_plasm{
           public:
//--current relative abundance, previous step value, previous iteration value, initial value--
            double x,x_,xi, begin;
            int param; //--number of component (technical variable)--
            double N, num_of_el; //--concentration (cgs), number of electrons which can be bind with this component--

            FILE *fout; //--file-pointer associated with component--

            comp_of_plasm *Self, *Previous, *Next; //--pointers for creation list of components--
            comp_of_plasm(int,double,double,char*);
            void Init_plasm_comp(double begin1);
            ~comp_of_plasm();
            void abundance(comp_of_plasm*);  //--function for calculation of current abundance--
            };

comp_of_plasm::comp_of_plasm(int param1, double begin1, double num_of_el1,
char* file_name)
{
     begin=begin1;
     x=begin1;
     x_=begin1;
     xi=1.e8;
     param=param1;
     N=N_nuclear(z)*begin1;
     num_of_el=num_of_el1;
     if((fout=fopen(file_name,"w"))==NULL)
     cout<<"error of creation of file "<<file_name<<"\n";
}

void comp_of_plasm::Init_plasm_comp(double begin1)
{
     begin=begin1;
     x=begin1;
     x_=begin1;
     xi=1.e8;
     N=N_nuclear(z)*begin1;
}

void comp_of_plasm::abundance(comp_of_plasm* first_comp) //--calculation of abundance of component--
{
     double var1=0.,var2=0.;
     comp_of_plasm *pointer=first_comp;
     while(pointer)
     {
             var1+=ww/H(z)*matrix[pointer->param][param]*pointer->x;
             var1+=ww_/H(z_)*(matrix_[pointer->param][param]*pointer->x_-matrix_[param][pointer->param]*x_);

             var2+=ww/H(z)*matrix[param][pointer->param];

	     pointer=pointer->Next;
     }
     var1+=-1./D_ln_1_plus_z*x_;
     var2+=-1./D_ln_1_plus_z;
     x=var1/var2;
}

comp_of_plasm::~comp_of_plasm() //--Destructor--
{
        fclose(fout);
}
  //--Class for ion component: end------------------------

  //--Class for electron component: begin-----------------

class electrons{
         public:
         double y, y_; //--relative abundance of free electrons--
	 double N; //--concentration of free electrons--
         FILE *fout; //--associated file-pointer--
	 electrons(char*);
         ~electrons();
	 void abundance(comp_of_plasm*); //--function for calculation of relative electron abundance
	 };

electrons::electrons(char *file_name)
{
     y=1.;
     N=N_e_tot(z)*y;
     if((fout=fopen(file_name,"w"))==NULL)
     cout<<"error of creation of file "<<file_name<<"\n";
}

void electrons::abundance(comp_of_plasm* first_comp)
{
     double res=0.;
     comp_of_plasm *pointer;
     pointer=first_comp;
     while(pointer)
     {
             res+=pointer->num_of_el*pointer->x;
	     pointer=pointer->Next;
     }
     y=delta_el*res;
}

electrons::~electrons()
{
   fclose(fout);
}
  //--Class for electron component: end-------------------
//--Class for description of plasma compenent: end---------------------

 comp_of_plasm HydI(0, 0., 0.,"HydI.dat"), HydII(1, 1., 1.,"HydII.dat"),
               HeI(2, 0., 0.,"HeI.dat"), HeII(3, 0., 1.,"HeII.dat"),
               HeIII(4, 0., 2.,"HeIII.dat");
 comp_of_plasm *first_comp=&HydI;

electrons El("el.dat");

void Set_pointers() //--Set pointers for creating list of plasma components--
{
//-----------------------Self------------------------------
       HydI.Self=&HydI;
       HydII.Self=&HydII;
       HeI.Self=&HeI;
       HeII.Self=&HeII;
       HeIII.Self=&HeIII;
//-----------------------Previous--------------------------
       HydI.Previous=0;
       HydII.Previous=&HydI;
       HeI.Previous=&HydII;
       HeII.Previous=&HeI;
       HeIII.Previous=&HeII;
//-----------------------Next------------------------------
       HydI.Next=&HydII;
       HydII.Next=&HeI;
       HeI.Next=&HeII;
       HeII.Next=&HeIII;
       HeIII.Next=0;
}


//--Special classes: begin------------------------------
  //--Class couple_double: begin--------------------
class couple_double
{
      public:   //couple of numbers (double) and pointer
      double arg;
      double fun;
      couple_double *Next;
      couple_double();
      void Init();
};

couple_double::couple_double()
{
      arg=0.;
      fun=0.;
      Next=0;
}

void couple_double::Init()
{
      arg=0.;
      fun=0.;
      Next=0;
}
    //--Class couple_double: end--------------------

    //--Class table_fun: begin--------------------
       //This class describes two-column table with dynamic length
class table_fun
{
      public:
      couple_double First;
      couple_double Fun_,Fun__;
      couple_double *i;
      double line_approx(double x);
      int In_table_fun(char *file_name,char header_sign, int, int);
      table_fun();
      ~table_fun();
};

table_fun::table_fun()
{
      i=&First;
}

table_fun::~table_fun()
{
     couple_double *pointer=First.Next;
     couple_double *pointer1;
     while(pointer)
     {
        pointer1=pointer->Next;
        delete pointer;
        pointer=pointer1;
     }
     First.Init();
     Fun_.Init();
     Fun__.Init();
     i=&First;
}

//--Function to obtain linear approximation from table data--
double table_fun::line_approx(double x)
{
  double res=0.;
  int signal=1;
     couple_double *pointer=&First;
     while((pointer->Next)&&signal)
     {
        if((x-pointer->arg)*(x-pointer->Next->arg)<=0.)
        {
        res=pointer->fun+(pointer->Next->fun-pointer->fun)/
        (pointer->Next->arg-pointer->arg)*(x-pointer->arg);
        signal=0;
        }
        pointer=pointer->Next;
     }
     return res;
}

//--Function for loading table data from data-file--
int table_fun::In_table_fun(char *file_name, char header_sign, int column_number,
int read_column)
{
     FILE *fout;
     if((fout=fopen(file_name,"rb+"))==NULL)
     {
        cout<<"error of opening file "<<file_name<<"\n";
        return 1;
     }
     couple_double *pointer=&First;
     double var;
	 int i;
     if(header_sign=='y')
     if(descriptor_shift(fout,'$'))
     {cout<<"message has been sent by"<<file_name<<"\n"; return 1;}

        if(fscanf(fout,"%lg",&(pointer->arg))==1)
        {
         for(i=2;i<read_column;i++)
         fscanf(fout,"%*lg");

         fscanf(fout,"%lg",&(pointer->fun));

         for(i=read_column+1;i<=column_number;i++)
         fscanf(fout,"%*lg");
        }
        while(fscanf(fout,"%lg",&var)==1)
        {
          pointer->Next=new couple_double;
          pointer=pointer->Next;
          pointer->arg=var;

          for(i=2;i<read_column;i++)
          fscanf(fout,"%*lg");

          fscanf(fout,"%lg",&(pointer->fun));

          for(i=read_column+1;i<=column_number;i++)
          fscanf(fout,"%*lg");
        }
     fclose(fout);
     return 0;
}
    //--Class table_fun: begin--------------------
//--Special classes: end------------------------------

//--Initialization of physical variables: begin------------------------

void init_general_assist_const()
{
alpha_fine=e*e/h/c; //--Calculation of fine structure constant--
me_reduced=me/(1.+me/mp);
I=me_reduced*e*e*e*e/h/h/2.; //--Calculation of hydrogen ionization energy--
freq1=I/2./pi/h; //-----------hydrogen ionization frequency---------
T_ion=I/kB; //--temperature equivalent of hydrogen ionization energy--
a_St_B=pi*pi*pow(kB,4)/60./c/c/h/h/h; //--Calculation of Stefan-Boltzmann constant--
sigma_t=(8.*pi/3)*(e*e/me/c/c)*(e*e/me/c/c); //--Calculation of Thomson cross section--
aB_2=pow(h,4)/pow(e,4)/pow(me,2); //--Calculation of square of Bohr radius--
mH=mp+me;

V_energy=pow(e/e_stand,4)*(me/me_stand)/pow(h/h_stand,2);
V_one_photon=pow(e/e_stand,10)*(me/me_stand)/pow(h/h_stand,6)/pow(c/c_stand,3);
V_two_photon=pow(e/e_stand,16)*(me/me_stand)/pow(h/h_stand,9)/pow(c/c_stand,6);
V_recombination=pow(e/e_stand,6)/pow(me/me_stand,1.5)/(h/h_stand)
                /pow(c/c_stand,3)/pow(kB/kB_stand,0.5);
}
//--Initialization of physical variables: end------------------------

//--Initialization of atomic data: begin-----------------------------

int HeII_atomic_data()
{
//--Energy levels of helium ion HeII--
I2s_HeII=I;
I1_HeII=4.*I;

//--two-photon transition probability for helium ion, s^{-1}---
A_2s1sHeII=pow(2.,6)*8.22458*V_two_photon;
//--obtained by scaling Z^6 from two-photon transition probability of hydrogen atom
//--see, e.g., Shapiro and Breit 1959, Zon and Rapoport 1968

//--transition wavelength, cm-------
lambd_aHeII=2.*pi*h*c/(I1_HeII-I2s_HeII); //--for Ly-aplha transition in helium ion-----

  return 0;
}

int HeI_atomic_data()
{
//--NIST Data are used--
//--http://physics.nist.gov/PhysRefData/ASD/lines_form.html--
//--http://physics.nist.gov/PhysRefData/ASD/levels_form.html--
//--Energy levels of helium atom HeI--
I1_HeI=24.5874*eV_erg*V_energy;

I2p_HeI=3.3694*eV_erg*V_energy;
I2s_HeI=3.9716*eV_erg*V_energy;
E_2p2sHeI=-(I2p_HeI-I2s_HeI);

I2p_orto_HeI=3.6233*eV_erg*V_energy;
I2s_orto_HeI=4.7678*eV_erg*V_energy;
E_2p2sHeI_orto=-(I2p_orto_HeI-I2s_orto_HeI);

//--Energy levels, Statistical weights, and Transition Probabilities--
//--Ortho Structure--
E_HeI_ortho_2s_1=19.8196134*V_energy;
g_HeI_ortho_2s_1=3;

//E_HeI_ortho_2p_0=20.9642176*V_energy;
//g_HeI_ortho_2p_0=1;
E_HeI_ortho_2p_1=20.9640951*V_energy;
g_HeI_ortho_2p_1=3;   
//E_HeI_ortho_2p_2=20.9640857*V_energy;
//g_HeI_ortho_2p_2=5;
g_HeI_ortho_2p=9; //--Statistical weight of ortho 2p level (taking into account all values of total moment J)

E_HeI_ortho_3s_1=22.7184651*V_energy;
g_HeI_ortho_3s_1=3;

E_HeI_ortho_3d_1=23.0736551*V_energy;
g_HeI_ortho_3d_1=3;
E_HeI_ortho_3d_2=23.0736496*V_energy;
g_HeI_ortho_3d_2=5;
//E_HeI_ortho_3d_3=23.0736493*V_energy;
//g_HeI_ortho_3d_3=7;

//A_HeI_ortho_2p_0_2s_1=1.022e7;
A_HeI_ortho_2p_1_2s_1=1.022e7;
//A_HeI_ortho_2p_2_2s_1=1.022e7;

//A_HeI_ortho_3s_1_2p_0=3.08e06;
A_HeI_ortho_3s_1_2p_1=9.25e06;
//A_HeI_ortho_3s_1_2p_2=1.54e07;

//A_HeI_ortho_3d_1_2p_0=3.92e07;
A_HeI_ortho_3d_2_2p_1=5.29e07;
A_HeI_ortho_3d_1_2p_1=2.94e07;
//A_HeI_ortho_3d_3_2p_2=7.06e07;
//A_HeI_ortho_3d_2_2p_2=1.76e07;
//A_HeI_ortho_3d_1_2p_2=1.96e06;

//--Para Structure--
E_HeI_para_1s_0=0.;
g_HeI_para_1s_0=1;

E_HeI_para_2s_0=20.6157736*V_energy;
g_HeI_para_2s_0=1;

E_HeI_para_2p_1=21.2180214*V_energy;
g_HeI_para_2p_1=3;

E_HeI_para_3s_0=22.9203161*V_energy;
g_HeI_para_3s_0=1;

E_HeI_para_3d_2=23.0740734*V_energy;
g_HeI_para_3d_2=5;

E_HeI_para_4s_0=23.6735693*V_energy;
g_HeI_para_4s_0=1;

E_HeI_para_4d_2=23.7363336*V_energy;
g_HeI_para_4d_2=5;

E_HeI_para_5s_0=24.0112136*V_energy;
g_HeI_para_5s_0=1;

E_HeI_para_5d_2=24.0428020*V_energy;
g_HeI_para_5d_2=5;

A_HeI_para_2p_1_1s_0=1.799e9;
A_HeI_para_2p_1_2s_0=1.976e06;
A_HeI_para_3s_0_2p_1=1.81e07;
A_HeI_para_3d_2_2p_1=6.38e07;
A_HeI_para_4s_0_2p_1=6.55e06;
A_HeI_para_4d_2_2p_1=2.02e07;
A_HeI_para_5s_0_2p_1=3.13e06;
A_HeI_para_5d_2_2p_1=9.07e06;

//--Probability of 2p->1s Ortho-Para transition, Lach and Pachucki, 2001---------------
A_HeI_ortho_2p_1_para_1s_0=177.58*V_one_photon; 

//--two-photon transition probability for helium atom, s^{-1}---
A_2s1sHeI=51.3*V_two_photon;
//--probability for semi-forbidden triplet-singlet dipole transition in helium atom, s^{-1}---
A_2p1s_ortoHeI=g_HeI_ortho_2p_1*A_HeI_ortho_2p_1_para_1s_0/g_HeI_ortho_2p;
//--probability for allowed singlet-singlet dipole transition in helium atom, s^{-1}---
A_2p1s_paraHeI=A_HeI_para_2p_1_1s_0;

//-- transition wavelength, cm-------
lambd_aHeI=2.*pi*h*c/(I1_HeI-I2p_HeI); //--for 2p1s transition in singlet structure of helium
lambd_a_ortoHeI=2.*pi*h*c/(I1_HeI-I2p_orto_HeI); //--for 2p1s transition in triplet structure of helium

    return 0;
}

int HydI_atomic_data()
{
//--Energy levels of hydrogen atom--
I2s_HydI=0.25*I;
I1_HydI=I;

//--two-photon transition probability for hydrogen atom, s^{-1}---
A_2s1sHydI=8.22458*V_two_photon;

//-- transition wavelength, cm-------
lambd_aHydI=2.*pi*h*c/(I1_HydI-I2s_HydI); //--for Ly-aplha transition in hydrogen atom--

   return 0;
}
//--Initialization of atomic data: end-----------------------------

//--Radiation functions: begin-----------------------------------------

double planckian_photon_per_mode(double freq, double T)
{
      return 1./(exp(2.*pi*h*freq/kB/T)-1.);
}
//--Radiation functions: end-------------------------------------------

//--Matter functions: begin--------------------------------------------

double g_e(double Te) //--partial function of free electrons----
{
        return pow((me*kB*Te/2./pi),3./2.)/h/h/h;
}

double T_matter(double z)
{
   static double T_decoup=0.;
   static double z_decoup=0.;
   double dz=D_ln_1_plus_z*(1.+z);
   double conc_var, conc_ratio;
   double res;
   double epsT1=0.;
   double F_T=8.*sigma_t*4.*a_St_B_stand*pow(T_br(z),4);
   F_T/=3.*me*c*c*H(z);
   conc_var=N_nuclear(z)*(HydII.begin+HeIII.begin)/N_e_tot(z);
   conc_ratio=El.y/(El.y + conc_var);
   F_T*=conc_ratio;
   epsT0=1./(F_T+1.);
   epsT1=(1.+z)*(epsT0-epsT0_)/dz/F_T;
   if(epsT0<=0.5) 
   {
      res=T_br(z)*(1.-(epsT0+epsT1));
      z_decoup=z;
      T_decoup=res;
   }
   else
   {
      res=T_decoup*pow((1.+z)/(1.+z_decoup),2);
   }
   return res;
}
//--Matter functions: end--------------------------------------------

//--Recombination coefficients: begin----------------------------------

//--Recombination coefficient for hydrogen--
//--Pequignot, Petitjean, and Boisson, 1991--
double recomb_HydII(double T)
{
    double aa=4.309, bb=-0.6166, cc=0.6703, dd=0.5300;
    double t=T/1.e4;
    return 1.e-13*aa*pow(t,bb)/(1.+cc*pow(t,dd))*V_recombination;
}

double recomb_HydII_fudged(double T)
{
//--with taking into account hydrogen fudge factor F_H--
    if(F_H<=0.) F_H=1.;
    return F_H*recomb_HydII(T);
}

//--Ionization coefficient for hydrogen using detailed balance--
/*double ion_HydI(double T)
{
     return recomb_HydII(T)*g_e(T)*exp(-I2s_HydI/kB/T);
}*/

double ion_HydI_fudged(double T)
{
     return recomb_HydII_fudged(T)*g_e(T)*exp(-I2s_HydI/kB/T);
}

//--Recombination coefficient to singlet states of helium atom--
//--Hummer and Storey, 1998--
double recomb_HeII(double T)
{
       double qq=pow(10.,-16.744), pp=0.711, T1=pow(10.,5.114), T2=3.;
       return 1.e6*qq/(sqrt(T/T2)*pow((1.+sqrt(T/T2)),(1.-pp))
       *pow((1.+sqrt(T/T1)),(1.+pp)))*V_recombination;
}

//--Recombination coefficient to triplet states of helium atom--
double recomb_orto_HeII(double T)
{
       double qq=pow(10.,-16.306), pp=0.761, T1=pow(10.,5.114), T2=3.;
       return 1.e6*qq/(sqrt(T/T2)*pow((1.+sqrt(T/T2)),(1.-pp))
       *pow((1.+sqrt(T/T1)),(1.+pp)))*V_recombination;
}

//--Ionization coefficient (using detailed balance)-- 
//--from singlet states of helium atom--
double ion_HeI(double T)
{
     return 4.*recomb_HeII(T)*g_e(T)*exp(-I2s_HeI/kB/T);
}

//--Ionization coefficient (using detailed balance)-- 
//--from triplet states of helium atom--
double ion_orto_HeI(double T)
{
     return 4./3.*recomb_orto_HeII(T)*g_e(T)*exp(-I2s_orto_HeI/kB/T);
}

//--Recombination coefficient of helium ion-- 
//--obtained by scaling from recombination coefficient for hydrogen--
//--see, e.g., Verner and Ferland 1996--
double recomb_HeIII(double T)
{
       return 2.*recomb_HydII(T/4.);
}

//--Ionization coefficient (using detailed balance) of HeII--
double ion_HeII(double T)
{
     return recomb_HeIII(T)*g_e(T)*exp(-I2s_HeII/kB/T);
}
//--Recombination coefficients: end----------------------------------

//--Hydrogen crossection required for calculation of 
//--continuum absorption during helium recombination: begin------------

//--Gaunt factor for hydrogen ionization cross section--
//--L.C. Johnson 1972-----
double gaunt_ion(double k,double freq)
{
          double var=freq1/k/k/freq;
	  double kk=1./k;
	  double g0,g1,g2;
      g0=1.133*(k==1)+1.0785*(k==2)+(k>=3)*
		  (0.9935+0.2328*kk-0.1296*kk*kk);
	  g1=-0.4059*(k==1)-0.2319*(k==2)-(k>=3)*
		  (0.6282-0.5598*kk+0.5299*kk*kk)*kk;
	  g2=0.07014*(k==1)+0.02947*(k==2)+(k>=3)*
		  (0.3887-1.181*kk+1.47*kk*kk)*kk*kk;
     return g0+g1*var+g2*var*var;
}

double hydrogen_ion_crs_sctn(double i,double freq)
{
        double res=me*pow(e,10)/pow(3.,1.5)/pi/pi/c/pow(h,6);
        res*=Hav(freq-freq1/i/i)/(pow(i,5)*pow(freq,3));
        res*=gaunt_ion(i,freq);
        return res;
}
//--Hydrogen crossection: end------------------------------------  

//--Functions described helium HeIII->HeII recombination: begin--------------

//--Effective coefficient of HeII 2p-1s radiative transitions, s^{-1}--
double A_r_heliumII(double z, double HeII_ground_state)
{
  return 8.*pi*H(z)/(3.*pow(lambd_aHeII,3)*N_nuclear(z)*(HeII_ground_state+epsilon0));
}
//--Functions described helium HeIII->HeII recombination: end----------------

//--Functions described helium HeII->HeI recombination: begin--------------
//--Relative Doppler line width for helium atoms and ions--
double relative_Dopler_line_width(double T)
{
    return sqrt(2.*kB*T/(He_to_H*mH))/c;
}

//--Calculation of single-scattering albedo for scatterings at HeI 1s-2p--
//--transitions in singlet structure and absorption coefficients--
double Gamma_HeI_para_2p(double z)
{
   double res;
   double E_2p1s=(E_HeI_para_2p_1-E_HeI_para_1s_0)*eV_erg;
   double E_2p2s=(E_HeI_para_2p_1-E_HeI_para_2s_0)*eV_erg;
   double E_3s2p=(E_HeI_para_3s_0-E_HeI_para_2p_1)*eV_erg;
   double E_3d2p=(E_HeI_para_3d_2-E_HeI_para_2p_1)*eV_erg;
   double E_4s2p=(E_HeI_para_4s_0-E_HeI_para_2p_1)*eV_erg;
   double E_4d2p=(E_HeI_para_4d_2-E_HeI_para_2p_1)*eV_erg;
   double E_5s2p=(E_HeI_para_5s_0-E_HeI_para_2p_1)*eV_erg;
   double E_5d2p=(E_HeI_para_5d_2-E_HeI_para_2p_1)*eV_erg;

    res=A_HeI_para_2p_1_1s_0*(1.+exp(-E_2p1s/kB/T_br(z)));
   res+=A_HeI_para_2p_1_2s_0*(1.+exp(-E_2p2s/kB/T_br(z)));
   res+=(g_HeI_para_3s_0/g_HeI_para_2p_1)*A_HeI_para_3s_0_2p_1*exp(-E_3s2p/kB/T_br(z));
   res+=(g_HeI_para_3d_2/g_HeI_para_2p_1)*A_HeI_para_3d_2_2p_1*exp(-E_3d2p/kB/T_br(z));
   res+=(g_HeI_para_4s_0/g_HeI_para_2p_1)*A_HeI_para_4s_0_2p_1*exp(-E_4s2p/kB/T_br(z));
   res+=(g_HeI_para_4d_2/g_HeI_para_2p_1)*A_HeI_para_4d_2_2p_1*exp(-E_4d2p/kB/T_br(z));
   res+=(g_HeI_para_5s_0/g_HeI_para_2p_1)*A_HeI_para_5s_0_2p_1*exp(-E_5s2p/kB/T_br(z));
   res+=(g_HeI_para_5d_2/g_HeI_para_2p_1)*A_HeI_para_5d_2_2p_1*exp(-E_5d2p/kB/T_br(z));
   return res;
}

double Gamma_HeI_orto_2p(double z)
{
   double res;
   double E_2p2s=(E_HeI_ortho_2p_1-E_HeI_ortho_2s_1)*eV_erg;
   double E_3s2p=(E_HeI_ortho_3s_1-E_HeI_ortho_2p_1)*eV_erg;
   double E_3d_2__2p=(E_HeI_ortho_3d_2-E_HeI_ortho_2p_1)*eV_erg;
   double E_3d_1__2p=(E_HeI_ortho_3d_1-E_HeI_ortho_2p_1)*eV_erg;

   res=A_HeI_ortho_2p_1_2s_1*(1.+exp(-E_2p2s/kB/T_br(z)));
   res+=(g_HeI_ortho_3s_1/g_HeI_ortho_2p_1)*A_HeI_ortho_3s_1_2p_1*exp(-E_3s2p/kB/T_br(z));
   res+=(g_HeI_ortho_3d_2/g_HeI_ortho_2p_1)*A_HeI_ortho_3d_2_2p_1*exp(-E_3d_2__2p/kB/T_br(z));
   res+=(g_HeI_ortho_3d_1/g_HeI_ortho_2p_1)*A_HeI_ortho_3d_1_2p_1*exp(-E_3d_1__2p/kB/T_br(z));
   return res;
}

//--Coefficient of absorption (cm^{-3}s^{-1}Hz^{-1})--
//--at HeI 1s-2p radiative transitions in singlet structure--
//--at the central frequency of singlet HeI 2p-1s line--
double kappa_He_para(double z, double HeI_ground_state)
{
    double freq_2p1spara=(I1_HeI-I2p_HeI)/2./pi/h;
    double Dfreq_D=freq_2p1spara*relative_Dopler_line_width(T_br(z));
    double b_f=Gamma_HeI_para_2p(z)/4./pi/Dfreq_D;
    return 3.*A_2p1s_paraHeI/sqrt(pi)/Dfreq_D*(HeI_ground_state+epsilon0);
}

//--Coefficient of absorption  (cm^{-3}s^{-1}Hz^{-1})--
//--at HeI 1s-2p radiative transitions in triplet structure--
//--at the central frequency of triplet HeI 2p-1s line--
double kappa_He_orto(double z, double HeI_ground_state)
{
    double freq_2p1sorto=(I1_HeI-I2p_orto_HeI)/2./pi/h;
    double Dfreq_D=freq_2p1sorto*relative_Dopler_line_width(T_br(z));
    double b_f=Gamma_HeI_orto_2p(z)/4./pi/Dfreq_D;
    return 9.*A_2p1s_ortoHeI/sqrt(pi)/Dfreq_D*(HeI_ground_state+epsilon0);
}

//--Coefficient of absorption  (cm^{-3}s^{-1}Hz^{-1})--
//--at HI 1s-c radiative ionizations--
double kappa_H(double z, double freq, double HI_ground_state)
{
   return hydrogen_ion_crs_sctn(1, freq)*8.*pi*freq*freq/c/c*(HI_ground_state+epsilon0);
}

//--Probability of transitions due to absorption of--
//--resonant quanta in continuum (taking into account--
//--Doppler profile of line only),--
//--Kholupenko, Ivanchik, and Varshalovich 2007, 2008--
double P_D(double gamma)
{
  double p,q;
  if(gamma<=5.e2)
  {
    p=0.66;
    q=0.9;
  }
  if((gamma>5.e2)&&(gamma<=5.e4))
  {
    p=0.515;
    q=0.94;
  }
  if((gamma>5.e4)&&(gamma<=5.e5))
  {
    p=0.416;
    q=0.96;
  }
  if((gamma>5.e5))
  {
    p=0.36;
    q=0.97;
  }
  return 1./(1.+p*pow(gamma,q));
}

//--Probability of transitions due to absorption of--
//--resonant quanta in continuum, taking into account--
//--Lorenz wings of Voigt profile at complete redistribution--
//--of resonant photons on frequency--
//--Kholupenko, Ivanchik, and Varshalovich 2007, 2008--
double P_W(double a, double gamma)
{
  double wing_factor=0.;
  double a1=1.25;
  double a2=1.e20;
  if(a)
  {
    wing_factor=2.*sqrt(a/gamma/pow(pi,1.5))*
    (atan(sqrt(a2*a2*sqrt(pi)/gamma/a))
    -atan(sqrt(-a1*a1*log(a)*sqrt(pi)/gamma/a)));
  }
  return wing_factor;
}

//--P_W1 is P_W modified for including into total probability--
//--at partial redistribution, Kholupenko, Ivanchik, and Varshalovich 2007, 2008--
//--It is contribution from "far" wings of Voigt profile where--
//--Raman scattering plays a crucial role in forming the emission profile--
double P_W1(double a, double gamma, double lambda)
{
  double wing_factor=0.;
  double a1=1.;
  double a2=1.e20;
  if(a)
  {
    if(gamma>=1.1)
    a1=0.5*sqrt(log(gamma));
    wing_factor=2.*sqrt(a/gamma/(1.-lambda)/pow(pi,1.5))*
    (atan(sqrt(a2*a2*sqrt(pi)/gamma/a/(1.-lambda)))
    -atan(a1*pow(2,0.25)*pow(pi,1./8.)*pow(lambda,0.25)*pow(gamma*a,-0.25)*pow((1.-lambda),-0.5)));
  }
  return (1.-lambda)*wing_factor;
}

//--Approximation of the following hypergeometric function--
//--F(0.5,2.,alpha+2.,0.5) for alpha>=0.25--
double hgf_approx(double alpha)
{
   double a=-1.07;
   double b=-0.45;
   return 1.+exp(a*log(alpha+1.5)+b);
}

//--Approximation for the following expression--
//--Gamma(alpha+1.5)/Gamma(alpha+2.) for alpha>=0.25--
double Gamma_ratio_appr(double alpha)
{
   return pow(alpha+1.28,-0.5);
}

//--Probability according to Chugai 1987 and Grachev 1988--
//--It is contribution to the total probability from "near"--
//--wings where frequency diffusion plays a crucial role --
//--in forming the emission profile--
double P_Ch_G(double lambda, double a, double gamma)
{
   double q=2.*sqrt(pi)/lambda/a/gamma;
   double alpha=(1.-lambda)/2./lambda/sqrt(q)-1./4.;
   double res=0.;
   if(a*gamma>=1.)
   res=1./sqrt(pi)*pow(8.*lambda*a/sqrt(pi),0.25)*pow(gamma, -0.75);
   res*=Gamma_ratio_appr(alpha);
   res*=hgf_approx(alpha);
   return res;
}

//--Total probability of destruction of resonant photons at --
//--continuum absorption. In considered case this is absorption--
//--at ionization of hydrogen--
double P_H(double gamma_f, double b_f, double lambda)
{
   double res=0.;
   double lambda_=1.-lambda;
   if(fabs(lambda_)<=1.e-2)
   {
      res=P_D(gamma_f)+P_Ch_G(lambda, b_f, gamma_f)+P_W1(b_f, gamma_f, lambda);
   }
   else
   {
      res=P_D(gamma_f)+P_W(b_f, gamma_f);
   }
   return res;
}

//--Effective coefficient of HeI 2p-1s radiative transitions in--
//--singlet structure due to escape of photons from line profile, s^{-1}--
double A_r_para_heliumI(double z, double HeI_ground_state)
{
  double res=8.*pi*H(z)/(3.*pow(lambd_aHeI,3)*N_nuclear(z)*(HeI_ground_state+epsilon0));
  double tau=A_2p1s_paraHeI/res;
  res*=(1-exp(-tau));
  return res;
}

//--Effective coefficient of HeI 2p-1s radiative transitions in--
//--triplet structure due to escape of photons from line profile, s^{-1}--
double A_r_ortho_heliumI(double z, double HeI_ground_state, double HI_ground_state)
{
  double gamma_f=kappa_He_orto(z, HeI_ground_state)/kappa_H(z,c/lambd_a_ortoHeI, HI_ground_state);
  double res=8.*pi*H(z)/(9.*pow(lambd_a_ortoHeI,3)*N_nuclear(z)*(HeI_ground_state+epsilon0));
  double tau=A_2p1s_ortoHeI/res;
  res*=(1-exp(-tau));
  res*=pow((1.+1./gamma_f),-2);
  return res;
}


//--Effective coefficient of HeI 2p-1s radiative transitions in--
//--singlet structure due to destruction of HeI 2p-1s resonant--
//--photons by neutral hydrogen, s^{-1}--
double A_H_para_heliumI(double z, double HeI_ground_state, double HI_ground_state)
{
    double freq_2p1spara=(I1_HeI-I2p_HeI)/2./pi/h;
    double gamma_f=kappa_He_para(z, HeI_ground_state)/kappa_H(z,freq_2p1spara, HI_ground_state);
    double Gamma_para_2p_var=Gamma_HeI_para_2p(z);
    double b_f=Gamma_para_2p_var/4./pi/relative_Dopler_line_width(T_br(z))/freq_2p1spara;
    double HeI_2p1s_para_albedo=A_HeI_para_2p_1_1s_0/Gamma_para_2p_var;
    return A_2p1s_paraHeI*P_H(gamma_f, b_f, HeI_2p1s_para_albedo);
}

//--Effective coefficient of HeI 2p-1s radiative transitions in--
//--triplet structure due to destruction of HeI 2p-1s resonant--
//--photons by neutral hydrogen, s^{-1}--
double A_H_ortho_heliumI(double z, double HeI_ground_state, double HI_ground_state)
{
    double freq_2p1sorto=(I1_HeI-I2p_orto_HeI)/2./pi/h;
    double gamma_f=kappa_He_orto(z, HeI_ground_state)/kappa_H(z,freq_2p1sorto, HI_ground_state);
    double Gamma_orto_2p_var=Gamma_HeI_orto_2p(z);
    double b_f=Gamma_orto_2p_var/4./pi/relative_Dopler_line_width(T_br(z))/freq_2p1sorto;
    double HeI_2p1s_orto_albedo=A_HeI_ortho_2p_1_para_1s_0/Gamma_orto_2p_var;
    return A_2p1s_ortoHeI*P_H(gamma_f, b_f, HeI_2p1s_orto_albedo);
}

//--Functions described helium HeII->HeI recombination: end-----------------

//--Functions described hydrogen recombination: begin-------------
//--Effective coefficient of HI 2p-1s radiative transitions due to--
//--escape of photons from line profile because of expansion of Universe, s^{-1}--
double A_r_hydrogen(double z, double HI_ground_state)
{
  return 8.*pi*H(z)/(3.*pow(lambd_aHydI,3)*N_nuclear(z)*(HI_ground_state+epsilon0));
}

//--Calculation of relative overheating of HI Ly-alpha radiation--
//--for calculating different radiative feedback effects--
//--Kholupenko, Ivanchik, and Varshalovich 2010--
double Hydrogen_radiation_overheating(double z)
{
    double res;
    double freq_2p1s=0.75*freq1;
    res=(El.y*N_e_tot(z)*recomb_HydII_var*HydII.x
        +(3.*A_r_hydrogen(z, HydI.x)+A_2s1sHydI)*planckian_photon_per_mode(freq_2p1s,T_br(z))*HydI.x);
    res/=(ion_HydI_var+A_2s1sHydI+3.*A_r_hydrogen(z, HydI.x))*HydI.x;
    res/=planckian_photon_per_mode(freq_2p1s,T_br(z));
    return (res-1.);
}

//--Out of ln of HI Ly-a relative overheating in dat-file--
//--for subsequent using at next iterations to calculate feedbacks--
int Out_of_Hydrogen_radiation_overheating_ln(FILE* fout)
{
    if(z<=2000.)
       {
        fprintf(fout,"%-20.8g",z);
        fprintf(fout,"%-20.8g",log(Hydrogen_radiation_overheating(z)));
        fprintf(fout,"\n");

       }
     return 0;
}

//--Declaration of table to store ln of relative HI Ly-a overheting--
//--as a function of redshift z--
table_fun ln_HydRO;

//--Initialization of table of ln of HI Ly-a overheting--
//--by data from dat-file--
int Init_HydRO()
{
     char filename[]="Hyd_Rad_Overheat.dat";
     return ln_HydRO.In_table_fun(filename, 'n', 2, 2);
}

//--Function for calculation of feedback inhibition coefficients C_n--
//--from ln of relative HI Ly-a overheting--
//--Kholupenko, Ivanchik, and Varshalovich 2010--
double C_Feedback_fun(int n, double z)
{   
    double res;
    double z_=((1.+z)*(1.-pow(((double)n+1),-2))/(1.-pow((double)n,-2))-1.);
    double var=ln_HydRO.line_approx(z);
    double var_=ln_HydRO.line_approx(z_);
    if((z<=1650.)&&var&&var_)
    res=(1.-exp(var_-var));
    else res=1.;
    return res;
}

//--Function for calculation of resonant feedback correction factor--
//--by which the "standard" HI 2p-1s transition rate should be multiplied--
//--Kholupenko, Ivanchik, and Varshalovich 2010--
double Hydrogen_FeedBack_Correction(double z, int step)
{
    double FeedBackCorrection=0.;
    if(step&&(Number_of_levels_for_Resonant_Feedback>=2))
    {
     double n_2;
     for(int n=3;n<=Number_of_levels_for_Resonant_Feedback;n++)
     {
       n_2=((double)n)*((double)n);
       FeedBackCorrection+=C_Feedback_fun(n, z)*pow((1.-1./n_2),3)*
       exp(-(1./4.-1./n_2)*I/kB/T_br(z));
     }
     FeedBackCorrection*=pow(4./3.,3);
     FeedBackCorrection+=C_Feedback_fun(2, z);
    }
    else
    FeedBackCorrection=1.;
    return FeedBackCorrection;
}

double Hydrogen_fudge_fun(double z)
{
    double var=(z-z_perturbation)/Delta_z_perturbation;
    return A_perturbation_H/(1.+var*var);
}

//--Functions described hydrogen recombination: end--------------

//--Calculating functions: begin-------------------------------
//--Procedure for storing solution of ODE from previous step--
void save_step()
{
    comp_of_plasm *pointer=first_comp;
    while(pointer)
     {
         pointer->x_=pointer->x;
	     pointer=pointer->Next;
     }
     El.y_=El.y;

     epsT0_=epsT0;
     T_matter_var_=T_matter_var;
}

int Set_redshifts(double ln_1_plus_z)
{
    z=redshift(ln_1_plus_z);
    z_=redshift(ln_1_plus_z-D_ln_1_plus_z);
    return 0;
}

//--Procedure for storing solution of nonlinear system--
//--from previous iteration in the frame of one redshift-step--
//--It is necessary for calculation of relative discrepancy--
//--of numerical solution--
void save_iter()
{
     comp_of_plasm *pointer=first_comp;
     while(pointer)
     {
      pointer->xi=pointer->x;
      pointer=pointer->Next;
     }
}

//--Procedure for setting variables xi to initial value--
void set_xi()
{
     comp_of_plasm *pointer=first_comp;
     while(pointer)
     {
         pointer->xi=1.e8;
         pointer=pointer->Next;
     }
}

//--Procedure for initialize matrix of reactions--
//--This is to avoid random settings matrix elements to non-zero values--
void Init_matrix()
{
      for(int i=0;i<quantity_speices;i++)
      for(int k=0;k<quantity_speices;k++)
        {
         matrix[i][k]=0.;
         matrix_[i][k]=0.;
        }
}

//--Initialization of reaction rate variables which are used--
//--instead of functions during one redshift-step--
//--Functions are not used directly to avoid unnecessary recalculating--
int Init_reaction_rates()
{
    recomb_HeIII_var=recomb_HeIII(T_br(z));
    ion_HeII_var=ion_HeII(T_br(z));

    recomb_HeII_var=recomb_HeII(T_br(z));
    ion_HeI_var=ion_HeI(T_br(z));

    recomb_orto_HeII_var=recomb_orto_HeII(T_br(z));
    ion_orto_HeI_var=ion_orto_HeI(T_br(z));

    recomb_HydII_var=recomb_HydII_fudged(T_matter_var);
    ion_HydI_var=ion_HydI_fudged(/*T_matter_var*/T_br(z));

//-----------------------------------------------------------------
    recomb_HeIII_var_=recomb_HeIII(T_br(z_));
    ion_HeII_var_=ion_HeII(T_br(z_));

    recomb_HeII_var_=recomb_HeII(T_br(z_));
    ion_HeI_var_=ion_HeI(T_br(z_));

    recomb_orto_HeII_var_=recomb_orto_HeII(T_br(z_));
    ion_orto_HeI_var_=ion_orto_HeI(T_br(z_));

    recomb_HydII_var_=recomb_HydII_fudged(T_matter_var_);
    ion_HydI_var_=ion_HydI_fudged(/*T_matter_var_*/T_br(z_));
//--------------------------------------------------------------
    return 0;
}

int Init_Additional_Kinetic_Coefficients(int step)
{
    Hydrogen_FeedBack_Correction_var=Hydrogen_FeedBack_Correction(z, step);
    return 0;
}

//---------------------------------------------------------
//--Saving Inhibition coefficients---------
int Save_Inhibition_Coefficients()
{
    CHeIII_=CHeIII; 
    CHeII_=CHeII; 
    CHeII_orto_=CHeII_orto; 
    CHydII_=CHydII;
    return 0;
}

//--Calculation of matrix of reaction rates--
void compute_matrix(int step)
{
    double T=T_br(z);
    double T_=T_br(z_);

    double Ne=N_e_tot(z);
    double Ne_=N_e_tot(z_);

       CHeIII=(3.*A_r_heliumII(z, HeII.x)+A_2s1sHeII)
              /(3.*A_r_heliumII(z, HeII.x)+A_2s1sHeII+ion_HeII_var);

//------------------------------------------------------------------

       double A_eff_Sum_para=3.*(A_H_para_heliumI(z, HeI.x, HydI.x)+A_r_para_heliumI(z, HeI.x))*exp(-E_2p2sHeI/kB/T)+A_2s1sHeI;
       CHeII=A_eff_Sum_para/(ion_HeI_var+A_eff_Sum_para);

       double A_eff_Sum_orto=3.*(A_H_ortho_heliumI(z, HeI.x, HydI.x)+A_r_ortho_heliumI(z, HeI.x, HydI.x))*exp(-E_2p2sHeI_orto/kB/T);
       CHeII_orto=A_eff_Sum_orto/(ion_orto_HeI_var+A_eff_Sum_orto);

//------------------------------------------------------------------------
       double A_eff_Sum_hydrogen=(3.*A_r_hydrogen(z, HydI.x)*Hydrogen_FeedBack_Correction_var+A_2s1sHydI);
       CHydII=A_eff_Sum_hydrogen/(A_eff_Sum_hydrogen+ion_HydI_var);

//---------------------HeIII----------------------------------------

       matrix[HeIII.param][HeII.param]=(CHeIII*El.y*Ne*recomb_HeIII_var);
       matrix[HeII.param][HeIII.param]=(CHeIII*ion_HeII_var*exp((I2s_HeII-I1_HeII)/kB/T));

//-----------------------HeII---------------------------------------

       matrix[HeII.param][HeI.param]=(CHeII*El.y*Ne*recomb_HeII_var
       +CHeII_orto*El.y*Ne*recomb_orto_HeII_var);

       matrix[HeI.param][HeII.param]=(CHeII*ion_HeI_var*exp((I2s_HeI-I1_HeI)/kB/T)
       +CHeII_orto*3.*ion_orto_HeI_var*exp((I2s_orto_HeI-I1_HeI)/kB/T));

//-----------------------HydII--------------------------------------

       matrix[HydII.param][HydI.param]=(CHydII*El.y*Ne*recomb_HydII_var);
       matrix[HydI.param][HydII.param]=(CHydII*ion_HydI_var*exp((I2s_HydI-I1_HydI)/kB/T/*T_matter_var*/));
//-------------------------------------------------------------------------
//------------------z_--------------------------------------------------
//---------------------HeIII----------------------------------------

       matrix_[HeIII.param][HeII.param]=(CHeIII_*El.y_*Ne_*recomb_HeIII_var_);
       matrix_[HeII.param][HeIII.param]=(CHeIII_*ion_HeII_var_*exp((I2s_HeII-I1_HeII)/kB/T_));

//-----------------------HeII---------------------------------------

       matrix_[HeII.param][HeI.param]=(CHeII_*El.y_*Ne_*recomb_HeII_var_
       +CHeII_orto_*El.y_*Ne_*recomb_orto_HeII_var_);

       matrix_[HeI.param][HeII.param]=(CHeII_*ion_HeI_var_*exp((I2s_HeI-I1_HeI)/kB/T_)
       +CHeII_orto_*3.*ion_orto_HeI_var_*exp((I2s_orto_HeI-I1_HeI)/kB/T_));

//-----------------------HydII--------------------------------------

       matrix_[HydII.param][HydI.param]=(CHydII_*El.y_*Ne_*recomb_HydII_var_);
       matrix_[HydI.param][HydII.param]=(CHydII_*ion_HydI_var_*exp((I2s_HydI-I1_HydI)/kB/T_/*T_matter_var_*/));
//------------------------------------------------------------------------------------

}

//--Calculation of abundances of plasma components--
void abundances(comp_of_plasm* first_comp)
{
     comp_of_plasm *pointer=first_comp;
     while(pointer)
     {
       pointer->abundance(first_comp);
       pointer=pointer->Next;
     }
     El.abundance(first_comp);
}

//--Calculation of relative discrepancy between solution--
//--at current iteration and solution from previous iteration--
double norm()
{
     double res=0.;
     comp_of_plasm *pointer=first_comp;
     while(pointer)
     {
         if(pointer->x)
         res+=fabs((pointer->x-pointer->xi)/pointer->x);
	     pointer=pointer->Next;
     }
     return res;
}

//--Calculating functions: end-------------------------------

//--Out data functions: begin--------------------------------
void streams_out_data(int mm)
{
      if(mm%number_of_info_points==0)
      {
        double Hydrogen_fudge_var=(1.+Hydrogen_fudge_fun(z));
         fprintf(El.fout,"%-20.8g",z);
         fprintf(El.fout,"%-20.8g",El.y*N0_e/N0_H*Hydrogen_fudge_var);
         fprintf(El.fout,"\n");

       comp_of_plasm *pointer=first_comp;
       while(pointer)
       {
         fprintf(pointer->fout,"%-20.8g",z);
         fprintf(pointer->fout,"%-20.8g",pointer->x);
         fprintf(pointer->fout,"\n");
	       pointer=pointer->Next;
       }
     }
}
//--Out data functions: end--------------------------------

//--Set Initial Conditions: begin------------------
int Set_Initial_Conditions(double z_begin)
{
     if(z_begin>=6090.)
     {
        HeIII.Init_plasm_comp(Y_N);
        HeII.Init_plasm_comp(0.);
     }
     else
     {
        HeIII.Init_plasm_comp(0.);
        HeII.Init_plasm_comp(Y_N);
     }
        HeI.Init_plasm_comp(0.);

        HydII.Init_plasm_comp(X_N);
        HydI.Init_plasm_comp(0.);

     T_matter_var=T_br(z_begin);
     T_matter_var_=T_br(z_begin);
     epsT0=0.;
     epsT0_=0.;
   
     return 0;
}
//--Set Initial Conditions: end--------------------

//--Iteration step function for calculation of ``fine''----
//--effects of recombination-------------------------------

int Iteration_of_Process(int step, int Number_of_Iterations)
{
  z=z_begin;
  Set_Initial_Conditions(z_begin);

  
    FILE* Hyd_Rad_Overheat_fout;

    if(step)
     Init_HydRO();

    if(!step)
        if((Hyd_Rad_Overheat_fout=fopen("Hyd_Rad_Overheat.dat","w"))==NULL)
        {
           cout<<"error of creation of Hyd_Rad_Overheat.dat\n";
        }

  int mm=0;
  Init_matrix();

  for(ln_1_plus_z=(log(1.+z_begin)+D_ln_1_plus_z);ln_1_plus_z>=log(1.+z_end);ln_1_plus_z+=D_ln_1_plus_z)
  {
//      int num_of_iter=0;
      Set_redshifts(ln_1_plus_z);
      set_xi();
      Init_Additional_Kinetic_Coefficients(step);
      while(norm()>norm_var)
      {
        Init_reaction_rates();
        compute_matrix(step);
        save_iter();
        abundances(first_comp);
        T_matter_var=T_matter(z);
//        num_of_iter++;
      }
      if(step==(Number_of_Iterations-1))
      streams_out_data(mm);
      save_step();
      Save_Inhibition_Coefficients();

      mm++;

    if(!step)
     Out_of_Hydrogen_radiation_overheating_ln(Hyd_Rad_Overheat_fout);
  }
    if(!step)
     fclose(Hyd_Rad_Overheat_fout);

   return 0;
}

//--Main-------------------------------------

int main()
{
  if(init_redshift_range()) return 0;
  if(init_fine_effects()) return 0;
  if(init_fund_const()) return 0;
  if(init_fudge_factors()) return 0;
  init_general_assist_const();
  if(init_cosmological_const()) return 0;

  HeII_atomic_data();
  HeI_atomic_data();
  HydI_atomic_data();

  Set_pointers();

  for(int step=0;step<=(Number_of_Iterations-1);step++)
  Iteration_of_Process(step,Number_of_Iterations);

  return 0;
}
